/**
 * Auswertung eines Magnetostriktiven Positions-Sensors
 * 
 * (c) January 2025 by Florian Bauer (azensepp)
 * 
 * Quartett-Scanner:
 *    Das Signal muss ein Quartett aus 2 Dubletten sein, ein Originales und ein Reflektiertes
 *    
 *    ACHTUNG:
 *    
 *    1. Der Quartett-Scanner benötigt zum Einphasen ein Quartett. Wenn er eingerastet (LOCK) ist,
 *    können auch einzelne Signale fehlen. Der Quartett-Scanner ordnet die Messungen dem jeweiligen Peak-Kanal zu und
 *    rekonstruiert ggf. fehlenden Signale. Vor dem Einrasten werden keine Werte ausgegeben. 
 *    
 *    2. Die Werte DT und TM müssen einmalig bestimmt und im Programm hinterlegt werden. 
 *    Zur Bestimmung wird der Printmode (mit p) aus und Debug-Mode (mit d) eingeschaltet
 *    Aus den Ausgaben:  VAL:<val> X1:<T1> X2:<T2> X3:<T3> X4:<T4>
 *    DT    Abstand der Peaks in einem Dublett: DT = <T2>->T1> = <T4>-<T3>
 *    TM    "Länge" des Drahtes <T3>=2*TM-<T2> bzw. <T4>=2*TM - <T1>
 *    
 *    ISSUES:
 *    - DT hängt von der Spulenposition ab
 *    - TM hängt von der Länge des Sensors ab
 *    - Vor dem Einrasten werden ggf. keine Werte ausgegeben
 *    - Es ist auf gute Abschirmung der Leitungen zur Spule zu achten
 *    - Wenn Sprünge oder Wellen auftreten, kann es sein, dass DT oder TM nicht richtig eingestellt sind
 *    
 *    CONFIGURATION:
 * Verstaerker:
 * 3 x Invertierend A=47 x 10 x 3
 * Kooplung Stufe 1 und 2: C= 175pF, R=10k
 * 
 * Komparator:
 *  R=20k, Feedback zu +: 68k (Feedback-Resistor für Hysterese um OSzillationen zu vermeiden)
 *  
 * Zeitmessung:  
 *  #define TIMER1MEAS 
 *    Timer1-ICP1
 * 
 * Draht:
 *  Ni 
 *  d= 0.8mm   (Durchmesser)
 *  l= 49.5cm  (Länge)
 *  In 7mm-PVC Rohr
 *  3 Lagerungs-Ringe, um Kontakt des Drahtes mit der Rohrwand zu verhindern
 *  
 * Klemmung:
 *  hart: Draht beidseitig => Quartett
 * 
 * Magnet:
 *  Ringmagnet 12mm
 *  
 * Anregung:
 *  U(exc)=18.8V - 20.5V
 *  C(exc)=400pF, R(exc)=2.2k
 * 
 * Spule:
 *  ca 715 Windungen
 * 
 * Gate-Monoflop: 
 *  C=2.2nF => 26us
 *  R=14.7k
 *  
 *    
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

// Maximale Anzahl der zu messenden Peaks
#define NMEAS 32

// 2 Kanäle werden unterstützt
// Beachte: DTx und TMx sind Konstanten; Falls ein Scanner anders ist, müssen diese Werte angepasst werden

// Abstand zweier Peaks in einem Dublett
#define DT0 345
#define DT1 345

// Abstand der mittleren Peaks P2 und P3
// entspricht der Länge des Drahtes:
// P3 = 2*TM - P2
#define TM0 2787
#define TM1 2787

// Genauigkeit für Dublett-Erkennung
#define EPS 45

// Maximal erlaubte Änderung der Laufzeit eines Peaks
// => Begrenzung der maximalen Bewegungsgeschwindigkeit des Magnet-Cursors
#define EPS2 80


int runmode=0;    // Motor an/aus
int printmode=1;  // Ausgabe an/aus
int debugmode=0;  // Debug-Ausgabe an/aus

typedef struct {
    long val;
    long lastval;
    int upd;
} Channel;

typedef struct {
  long dt,tm;
  Channel * channels;
  int astate=0;
  long value;
} Measurement;

void backup(Channel channels[]) {
    for (int i = 0; i < 4; i++) {
        channels[i].lastval = channels[i].val;
        channels[i].upd = 0;
    }
}

void printc(Channel channels[]) {
    for (int i = 0; i < 4; i++) {
      Serial.print(channels[i].val);
      Serial.print(", ");
    }
    Serial.println();
}



Channel channels0[4] = {{0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0}};
Channel channels1[4] = {{0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0}};
Measurement meas[2];

void setupMeasurement(Measurement *meas0, Channel * channels0,long dt, long tm) {

    meas0->astate = 0;
    meas0->channels = channels0;
    meas0->dt=dt;
    meas0->tm=tm;
}

/*
 * Quartett-Tracker
 */
int analyze2( long * signal, Measurement *meas)
{
      int signal_length;
      long value=0;
      int i;
        if(debugmode==1)
        {
          Serial.print("SIG: ");
          Serial.print(meas->astate);
          Serial.print(" ");
        }
        for(i=0; i<NMEAS && signal[i]> 0 && signal[i]<10000;  i++)
        {
          if(debugmode==1) {
            Serial.print(signal[i]);
            Serial.print(" ");
          }
        }
        if(debugmode==1) Serial.println();
        
        signal_length=i;

        if (meas->astate == 0 && signal_length >= 4) { // Capture mode
            for (int i = 0; i < signal_length; i++) {
                for (int j = i + 1; j < signal_length; j++) {
                    if (signal[i] > 0 && signal[j] > 0) {
                        if (abs(signal[j] - signal[i] - meas->dt) < EPS) {
                            if (signal[i] < meas->tm && signal[j] < meas->tm) {
                                meas->channels[0].val = signal[i];
                                meas->channels[1].val = signal[j];
                                meas->astate = 1;
                            }
                            if (signal[i] >= meas->tm && signal[j] >= meas->tm && abs(2 * meas->tm - meas->channels[0].val - signal[j]) < EPS) {
                                 meas->channels[2].val = signal[i];
                                 meas->channels[3].val = signal[j];
                                if (meas->astate == 1) {
                                    meas->astate = 2;
                                  //printf("Locked at %d\n",t);
                                }
                            }
                        }
                    }
                }
            }
            if (meas->astate != 2) {
                for (int i = 0; i < 4; i++) {
                    meas->channels[i].val = 0;
                }
                meas->astate = 0;
            }
//            if (meas->astate == 2) {
//               printf("Locked: %d, ", t);
//               printc(channels);
 //           }
        } else if (meas->astate == 2) { // Tracking mode
            backup(meas->channels);
            int k = -1;
            for (int j = 0; j < 4; j++) {
                long delta=10000;
                // Assign best matching signals to channels
                for (int i = 0; i < signal_length && signal[i] !=0; i++) {
                    if (abs(meas->channels[j].lastval - signal[i]) < EPS2) {
                        int found=0;
                        
                        // avoid double assignment
                        for(int p=i-1; p>0; p--)
                        {
                          if( meas->channels[p].val==signal[i] )
                          {
                            found=1;
                          }
                        }
                        if( found==0 ) 
                        {
                          if((i<2 && signal[i]<meas->tm) ||(i>1 && signal[i]>=meas->tm)) { 
                            if( meas->channels[j].upd==0) 
                            {
                              meas->channels[j].val = signal[i]; // channel not updated => assign Signal value
                              meas->channels[j].upd = 1;         // mark updated
                              k = j;
                            }
                            else // channel already updated => assign Signal if it is nearer to last value:
                            {
                              long ndelta = abs(meas->channels[j].lastval - signal[i]);
                              if( ndelta < delta) // signal is closer to last value
                              {
                                meas->channels[j].val = signal[i]; // assign signal value
                                meas->channels[j].upd = 1;
                                k = j;
                                delta=ndelta;  // this is now the best assignment
                              }
                            }
                            
                          }
                        }
                       
                    }
                }
            }
            if (k >= 0) {
              // Reconstruct the missign signals:
                for (int j = 0; j < 4; j++) {
                    if (j != k) {
                        if (k == 0) {
                            if (j == 1 && meas->channels[j].upd == 0) {
                                meas->channels[j].val = meas->channels[k].val + meas->dt;
                            }
                            if (j == 2 && meas->channels[j].upd == 0) {
                                meas->channels[j].val = 2 * meas->tm - meas->channels[k].val - meas->dt;
                            }
                            if (j == 3 &&  meas->channels[j].upd == 0) {
                                meas->channels[j].val = 2 * meas->tm - meas->channels[k].val;
                            }
                        } else if (k == 1) {
                            if (j == 0 &&  meas->channels[j].upd == 0) {
                                meas->channels[j].val = meas->channels[k].val - meas->dt;
                            }
                            if (j == 2 &&  meas->channels[j].upd == 0) {
                                meas->channels[j].val = 2 * meas->tm - meas->channels[k].val;
                            }
                            if (j == 3 &&  meas->channels[j].upd == 0) {
                                meas->channels[j].val = 2 * meas->tm - meas->channels[k].val + meas->dt;
                            }
                        } else if (k == 2) {
                            if (j == 0 && meas->channels[j].upd == 0) {
                                meas->channels[j].val = 2 * meas->tm - meas->channels[k].val - meas->dt;
                            }
                            if (j == 1 && meas->channels[j].upd == 0) {
                                meas->channels[j].val = 2 * meas->tm - meas->channels[k].val;
                            }
                            if (j == 3 && meas->channels[j].upd == 0) {
                                meas->channels[j].val = meas->channels[k].val + meas->dt;
                            }
                        } else if (k == 3) {
                            if (j == 0 && meas->channels[j].upd == 0) {
                                meas->channels[j].val = 2 * meas->tm - meas->channels[k].val;
                            }
                            if (j == 1 &&  meas->channels[j].upd == 0) {
                                meas->channels[j].val = 2 * meas->tm - meas->channels[k].val + meas->dt;
                            }
                            if (j == 2 &&  meas->channels[j].upd == 0) {
                                meas->channels[j].val = meas->channels[k].val - meas->dt;
                            }
                        }
                    }
                }
               
            }
            else
            {
             // astate=0;
            }
        }
        if(debugmode==1) printc(meas->channels);
        
    value= meas->channels[1].val;
    /**
     * Fallback für unlocked
     */
    if(meas->astate==2 || meas->channels[1].upd)
    {
      value= meas->channels[1].val;
    }
    else if(meas->channels[2].upd)
    {
      value = 2*meas->tm-meas->channels[2].val;
    }
    else if(meas->channels[3].upd)
    {
     value= 2*meas->tm-meas->channels[3].val -meas->dt;
    }
    else if(meas->channels[0].upd)
    {
     value=meas->channels[0].upd+meas->dt;
    }     
    meas->value=value; 
    return value;
}


//#define MOTORSHIELD

#define MOTORSHIELD_V2

#ifdef MOTORSHIELD
//MOTOR A PINS  (LEFTMOTOR)
#define MOTOR_A_DIR_PIN 12
#define MOTOR_A_PWM_PIN 3
#define MOTOR_A_BRAKE_PIN 9
//#define MOTOR_A_CS_PIN A0
//MOTOR B PINS   (RIGHT MOTOR)
#define MOTOR_B_DIR_PIN 13
#define MOTOR_B_PWM_PIN 11
#define MOTOR_B_BRAKE_PIN 8
//#define MOTOR_B_CS_PIN A1
#endif

#ifdef MOTORSHIELD_V2
#include <Adafruit_MotorShield.h>
Adafruit_MotorShield AFMS = Adafruit_MotorShield();
// Select which 'port' M1, M2, M3 or M4. In this case, M1
Adafruit_DCMotor *myMotor[4] = { AFMS.getMotor(1), AFMS.getMotor(2), AFMS.getMotor(3), AFMS.getMotor(4) };

#endif

/*
 * PID controller for Motor position control
 */
//double kp = 1.8, ki = 1.8, kd = .00; //0.01; 
//double kp = 24, ki = 0.03, kd = 0; //0.01; 
//double kp = 35, ki = 0.03, kd = 0; //0.01; 



#ifdef DIFFERENTIAL
double kp = 50, ki = 0.3, kd = 1; //0.01; 
int llimit = 65;
int ulimit = 480;
#else
double kp = 1.5, ki = 45, kd = .005; //0.01; 
int llimit = 75;
int ulimit = 220;
#endif

// PID-Controller for Position Control:
#include <PID_v1_bc.h>
//double kp = 2, ki = .01, kd = .01; //0.01; 

double input1Old0=0, input0 = 0, output0 = 0, setpoint0 = 0, input0_0 = 0;
PID myPID0  =   PID (&input0, &output0, &setpoint0, kp, ki, kd, DIRECT);  
double input1Old1=0, input1 = 0, output1 = 0, setpoint1 = 0, input1_0 = 0;
PID myPID1  =   PID (&input1, &output1, &setpoint1, kp, ki, kd, DIRECT);  

// Option 1: Time Measurement with TTL-Counter:
// #define TTLCOUNTER
 /**
  * Option 1: Measurement with TTL-Counter
  *  
  * requires an external synchronous counter circuit clocked with 16 MHz 
  * TTL-Counter 
  *  1 x 16 MHz Quarz Generator 
  *  1 x 74LS08 AND-Gate
  *  1 x 74LS74 D-Flip-Flop
  *   D=H, R=inv(Arduino SEL2,Sample), CLK=inv(Signal)
  *  1 x 74LS04 Inverter
  *  3 x 74LS161  
  *  2 x 74LS153 4:1-Multiplexer
  *      MUX00  MUX01  MUX10    MUX11    A   B
  * C0     D0    D1      D2      D3      L   L
  * C1     D4    D5      D6      D7      L   H
  * C2     D8    D9      D10     D11     H   L
  *(C3     D12   D13     D14     D15     H   H) - only needed for 16 Bit counter
  *count range 0-2047 (4095 with 4 x '161)
  * 
  * Arduino
  * D4    SEL1       - Excitation - to Exciter (Active High)
  * D5    SEL2       - Sample    - to Counter Gate (Active High)
  * D6    CTRA       - Counter Nibble Select  
  * D7    CTRB       - Counter Nibble Select
  * A2-A5 COUNTER    - Counter Nibble data
  * D10   CHSEL      - Channel / Axis select
  * 
  * Counter:
  * SIG    - digital signal pulse from Comparator, stops counter (Active Low)
  * Sample - Signal from Arduino (Active High) - resets counter, measurement will start after Signal goes low
  */

// count range 0-2047 (4095 with 4 x '161)
//
#ifdef TTLCOUNTER
// Multiplexer-Inputs (Nibble Select): 
#define CTRA  6
#define CTRB  7

//Multiplexer-Outputs 
/*#define CTR0 A2
#define CTR1 A3
#define CTR2 A4
#define CTR3 A5
*/
#define CTR0 A0
#define CTR1 A1
#define CTR2 A2
#define CTR3 A3

#endif
#define SEL1 4  // Excitation
#define SEL2 5  // Sample
// Channel Select Pin 
#define CHSEL 9
//10

#ifdef TTLCOUNTER
/*
 * Select Nibble
 */
void writeCounterAdr(int  adr)
{
  switch(adr)
  {
    case 0: digitalWrite(CTRA,LOW); digitalWrite(CTRB,LOW); break;
    case 1: digitalWrite(CTRA,LOW); digitalWrite(CTRB,HIGH); break;
    case 2: digitalWrite(CTRA,HIGH); digitalWrite(CTRB,LOW); break;
    case 3: digitalWrite(CTRA,HIGH); digitalWrite(CTRB,HIGH); break;
  }
}
/*
 * read Nibble from counter
 */
int readCounterNibble(int adr)
{
  int x = 0;
  writeCounterAdr(adr);
  delayMicroseconds(2);
  if(digitalRead(CTR0)) x |= 1;
  if(digitalRead(CTR1)) x |= 2;
  if(digitalRead(CTR2)) x |= 4;
  if(digitalRead(CTR3)) x |= 8;
 
  return x;
}


/*
 * read Counter value
 */
int readCounter(int ch)
{
  if(ch==0) digitalWrite(CHSEL,0); else digitalWrite(CHSEL,1);
  
  int c = 0;
  c = readCounterNibble(0)   | (readCounterNibble(1) << 4 ) | (readCounterNibble(2) << 8);
  return c;
}

/*
 * Perform time measurement on Channel ch = 0 | 1 
 * with digital counter
 */
long measure_c(int ch)
{
     digitalWrite(CHSEL,ch);
     delayMicroseconds(1);

     // Emit pulses:
     digitalWrite(SEL1,HIGH); // Excitation Pulse
     digitalWrite(SEL2,HIGH);  // Sample Pulse
     delayMicroseconds(3);
     digitalWrite(SEL1,LOW);
    
     delayMicroseconds(30); // 30
  
     digitalWrite(SEL2,LOW);

     delayMicroseconds(250);

     return readCounter(ch);
    
  }

#endif


#define TIMER1MEAS
#ifdef TIMER1MEAS



long value_old=0;

volatile int count;
volatile int MeasEnd = 0;
//uint16_t 
long time,T_old, T0, T1, T2, T1OVF_Counter = 0;
//unsigned 
long T, Freq = 0;

// Option 2: Time Measurement with TTL-Counter:
volatile long Ts_old[NMEAS];
volatile long Ts[NMEAS];
#define CTL  10
//
// Time Measurement with TIMER1 and Input Capture on Pin 8
// Signal containing negative Start and Stop pulses (High->Low)
// --     ----------------------------  ---------
//  |     |                          |  |
//  -------                          ----
//  Start                            Stop
// (This Signal can be generated with an AND-Gate combining
// Trigger Pulse and Signal Puls
//
// Time measurement is performed between Start and Stop pulses
// Caveat: Not compatible with MOTORSHIELD - Motor B

// Option 3: Time Measurement with TTL-Counter:
// Time Measurement with IRQ on Pin 2  and micros() 
//#define IRQMEAS
/*
 * Setup Timer1 for Input Capture 
 */
void setupTimer1()
{
 
  // Input Capture setup
  // ICNC1: Enable Input Capture Noise Canceler
  // ICES1: =1 for trigger on rising edge
  // CS10: =1 set prescaler to 1x system clock (F_CPU)
  TCCR1A = 0;
  TCCR1B = (0<<ICNC1) | (0<<ICES1) | (1<<CS10);
  TCCR1C = 0;

  //catchFallingEdge(); // initialize to catch
   TCCR1B &= ~(1<<ICES1); TIFR1 |= (1<<ICF1);
  // Interrupt setup
  // ICIE1: Input capture
  // TOIE1: Timer1 overflow
  TIFR1 = (1<<ICF1) | (1<<TOV1);        // clear pending
  TIMSK1 = (1<<ICIE1) | (1<<TOIE1); // and enable
  //TIMSK1 |= B00100001;  // Enable Timer OVF & CAPT Interrupts
  
  // Set up the Input Capture pin, ICP1, which corresponds to Arduino D8
  pinMode(8, INPUT_PULLUP);
  //digitalWrite(8, 0);       // leave floating to count 60 Hz etc.
  //digitalWrite(8, 1);         // or enable the pullup


}
/*
 * Interrupt to count Overflows
 */
ISR(TIMER1_OVF_vect)
{
  T1OVF_Counter++;
  //T1OVF_Counter += 65536;
  //T1 -= 65536;
}

 /*
  * Interrupt for Pin Change at Pin 8 
  */
ISR(TIMER1_CAPT_vect)
{
   long T;
   if(count==0)
   {
     Ts[count]= (long)ICR1 + (65536 * T1OVF_Counter); ;
     if(count <NMEAS) count++; else MeasEnd=0;
     //digitalWrite(CTL,HIGH);delayMicroseconds(10); digitalWrite(CTL,LOW);
   }
   else 
   {
 //  digitalWrite(SEL3,HIGH);
      if(count>0)
      {
          T= (long)ICR1 + (65536 * T1OVF_Counter) - Ts[0]; 
          /*if(count <NMEAS)
          {
            if( Ts_new !=   Ts[count-1]) 
            {Ts[count] =Ts_new; count++; } 
          } else MeasEnd=0;
          */
          if(count <NMEAS)
            if(T !=   Ts[count-1]) {Ts[count] =T;  T_old=T; count++; } 
          else MeasEnd=0;

         //digitalWrite(CTL,HIGH);delayMicroseconds(20); digitalWrite(CTL,LOW);
      }
      if(count==NMEAS-1)     cli();
      //MeasEnd = 0; 
     
     
    } 
   
    
}

#define X_COIL 215L
#define X_MID 2590
#define X_LEN 5166
/*
 * read Counter value
 * heuristischer Analyzer, der bis 10cm vor der Spule funktioniert
 */
long measure_h0(int ch)
{
  long value=0;
  long x_coil = X_COIL;
  long x_mid  = X_MID; //2630; // 2851;
  long x_len  = X_LEN;
  
  int fall=0;
ch=1; //!!!!!!!!!!!!!!!!!!!!!!!!!!
  if(ch==0) 
    digitalWrite(CHSEL,0); 
  else 
    digitalWrite(CHSEL,1);
  
  do {
     T1OVF_Counter = 0;  
     count=0;     // Reset Puls Counter
     MeasEnd=1;   
     time = micros();
     sei();       // Enable Interrupt
     // Emit pulses:
     digitalWrite(SEL1,HIGH); // Excitation Pulse
     digitalWrite(SEL2,HIGH);  // Sample Pulse
     delayMicroseconds(1);
     digitalWrite(SEL1,LOW);
    
     delayMicroseconds(10); // 30
  
     digitalWrite(SEL2,LOW);

     //delayMicroseconds(250);

   // Wait for End of Measure
     while(micros()-time < 1000)
     {
        delayMicroseconds(10); 
     } 
//#define DBG
//#define DBG2
long t1,t2,t3,t4,t5;
    noInterrupts();
    t1=Ts[1]; t2=Ts[2]; t3=Ts[3]; t4=Ts[4]; t5=Ts[5]; // wegen Ueberschreiber in Interrupt-Routine !!
    if(t1<0) t1=0;
     if(t2<0) t2=0; if(t3<0) t3=0; if(t4<0) t4=0; if(t5<0) t5=0;
    interrupts();


    
    if(count<=4  && ((int)(abs( t4-t3 -x_coil) < 10)))
    {
      if(t4<x_mid)
      {
        value=t4; fall=100;
#ifdef DBG
       Serial.print(" 43a "); Serial.print(value); Serial.print(" "); 
#endif
      }
      else
      {
        if( t3>x_mid)
        {
          value=t2;
        }
        
        if( t1>x_mid && t2 > x_mid  && ((int)(abs( t2-t1 -x_coil) < 10)))
        {
                  value=2*x_mid-t1; fall=114;
        }
        else
        {
          
                  value=2*x_mid-t3; fall=110;
                if(value < 400)
                value=t2;
        }
        
#ifdef DBG
        Serial.print(" 43b "); Serial.print(value); Serial.print(" "); 
#endif
      }  
    }

    else if( count>=2 &&  ((int) (abs(t2-t1 -x_coil)) < 10))
    {
      if(t2<x_mid)
      {
      
        if(count>=5 && ((int) (abs(t5-t4 -x_coil)) < 10))
        {
          value = 2*x_mid - t4;
        }
        else
        {
          if(t2>x_mid)
          {
           value=2*x_mid-t2; 
          }
          else
          {
            value=t2; fall=100;
          }
        }
#ifdef DBG
        Serial.print(" 12a "); Serial.print(value); Serial.print(" "); 
#endif
      }
      else
      {
        value=2*x_mid-t2; fall=110;
#ifdef DBG
        Serial.print(" 12b "); Serial.print(value); Serial.print(" "); 
#endif
      }
    } else if( count>=3  && ((int)abs( t3- t2 -x_coil) < 10 ))
    {
      if(t2<x_mid)
      {
        value=t2; fall=100;
        if(t2<0)
        {
          if(count>=5 &&  ((int)abs( t5- t4 -x_coil) < 10 ))
          {
            value=x_mid*2 -t4;
          }
        }
#ifdef DBG
        Serial.print(" 32a "); Serial.print(value); Serial.print(" "); 
#endif
      }
      else
      {
        value=2*x_mid-t2; fall=110;
#ifdef DBG
        Serial.print(" 32b "); Serial.print(2*x_mid-t2); Serial.print(" "); 
#endif
      }
   }
   if(value==0)
   {
      if(t1<x_mid && t2>x_mid)
      {
        value=t1;
      }
else
      if(t2>x_mid)
      {
         value=2*x_mid-t2;
      }
      else
      {
         value=t2;
      }
   }
#ifndef DBG
    
  // Serial.print("X:");Serial.print(value);Serial.print(" Fall:"); Serial.println(fall);
    #endif
  
    
   for( int i=1;i<count;i++)
    {
#ifdef DBG
      Serial.print(" X"); Serial.print(i); Serial.print(":"); // Serial.print(Ts[i]);
      switch(i)
      {

        // use saved values insead of array:
        case 1: Serial.print(t1); break;
        case 2: Serial.print(t2); break;
        case 3: Serial.print(t3); break;
        case 4: Serial.print(t4); break;
        case 5: Serial.print(t5); break;
        default: Serial.print(Ts[i]); break;
      }
#endif
    }
#ifdef DBG  
    Serial.println();
#endif
  }
while(value==0);
  return value;
}


double valold=0;
/*
 * read Counter value
 * heuristischer analyzer, der spiket
 */
long measure_h(int ch)
{
  long value=0;
  long x_coil = X_COIL;
  long x_mid  = X_MID; // 2851;
  long x_len  = 0;
  
  int fall=0;
ch=1; //!!!!!!!!!!!!!!!!!!!!!!!!!!
  if(ch==0) 
    digitalWrite(CHSEL,0); 
  else 
    digitalWrite(CHSEL,1);
  
  do {
     T1OVF_Counter = 0;  
     for(int i=0;i<6;i++) Ts[i]=0;
     count=0;     // Reset Puls Counter
     MeasEnd=1;   
     time = micros();
     sei();       // Enable Interrupt
     // Emit pulses:
     digitalWrite(SEL1,HIGH); // Excitation Pulse
     digitalWrite(SEL2,HIGH);  // Sample Pulse
     delayMicroseconds(1);
     digitalWrite(SEL1,LOW);
    
     delayMicroseconds(40); // 30
  
     digitalWrite(SEL2,LOW);

     //delayMicroseconds(250);

   // Wait for End of Measure
     while(micros()-time < 1000)
     {
        delayMicroseconds(10); 
     } 

//#define DBG2
long T[6];

noInterrupts();

for(int i=0;i<6;i++) 
{
  T[i]=Ts[i];
  if(T[i]<0 || T[i] > 10000 ) T[i]=0;
}
interrupts();

int looping=1;
int mode=0;
/**
 * T1
 * T1 T2
 * T1 T2 T3 (Fehlpuls in der Mitte)
 */
// T[1] T[2] T[3]
if( count==2)
{
  if(T[1]>4000)
  {
    value=2*x_mid-T[1]; mode=1;
  }
  else
  {
    value=T[1]; mode=2;
  }
}
else if( count==3)
{
  if(T[2]>x_mid)
  {
    value=2*x_mid-T[2]; mode=3;
  }
  else if(T[1]>x_mid)
  {
    value=2*x_mid-T[1]; mode=4;
  }
  else
  {
    value=T[1];
  }
}
else 
 {
  int j=count-1;
  int looping=1;
  mode=20;
  value=T[1];
  do{
    if( T[j]>x_mid)
    {
         value=2*x_mid-T[j]; mode=10+j;
         if( abs(T[j]-T[j-1]-x_coil) > 50)
          {
         looping=0; 
         }
    } 
    j--;
  } while(j>0 && looping==1);
 }
 if(looping==0) { value=T[1]; mode=30; }
 
if(debugmode==1)
{   Serial.print(" VAL:");
    Serial.print(value);
    Serial.print(" M:");
    Serial.print(mode);

    for( int i=1;i<count;i++)
    {
      Serial.print(" X"); Serial.print(i); Serial.print(":"); Serial.print(T[i]);
    }
    Serial.println();
} 
}
 

while(value==0);
  valold = 0.333 * (double) value + 0.666* valold;
  value=valold;
  return value;
}
/*
 * read Counter value
 * heuristischer Analyzer, der einigermaßen funktioniert
 */
long analyze0(long * T, Measurement *  meas)
{
  long value=0;
  long x_coil = X_COIL;
  long x_mid  = X_MID; // 2851;
  long x_len  = 0;
  
  int fall=0;

  
int looping=1;
int mode=0;
long t00,t01;
int state=0;
for(int i=1;  i < count && state !=10; i++)
{
  //Serial.print("X"); Serial.print(i); Serial.print(":"); Serial.println(Ts[i]);
  if(state==0)
  {
    if( Ts[i]>300) // !!!! Ein hartcodierter Wert !!!!
    {
      if( i==count-1)
      {
        if( Ts[i]>4000) // !!!! Ein hartcodierter Wert !!!!
        {
           value=2*x_mid - Ts[i]; mode=5; state=10; 
        }
        else
        {
        state=10; value=Ts[i]; mode=1;
        }
      }
      else
      {
        t00=Ts[i]; state=1; mode=2;
      }
    }
  }
  else if(state==1)
  {
    t01=Ts[i];
    if( abs(t01-t00-meas->dt)<EPS2/2)
    {
      value = t01; mode=3;
      state=10;
    } else
    {
      t00=t01; state=1; mode=4;
    }
  }

}


 
if(debugmode==1)
{   Serial.print(" VAL:");
    Serial.print(value);
    Serial.print(" M:");
    Serial.print(mode);

    for( int i=1;i<count;i++)
    {
      Serial.print(" X"); Serial.print(i); Serial.print(":"); Serial.print(T[i]);
    }
    Serial.println();
} 

  return value;
}

/*
 * read Counter value
 * estimator mit analyze
 */
long measure(int ch)
{
  long value=0;
  long x_coil = X_COIL;
  long x_mid  = X_MID; // 2851;
  long x_len  = 0;
  
  int fall=0;
//if(ch==1) return 100;
ch=1; //!!!!!!!!!!!!!!!!!!!!!!!!!!

  if(ch==0) 
    digitalWrite(CHSEL,0); 
  else 
    digitalWrite(CHSEL,1);
  
  do {
     T1OVF_Counter = 0;  
     for(int i=0;i<6;i++) Ts[i]=0;
     count=0;     // Reset Puls Counter
     MeasEnd=1;   
     time = micros();
     sei();       // Enable Interrupt
     // Emit pulses:
     digitalWrite(SEL1,HIGH); // Excitation Pulse
     digitalWrite(SEL2,HIGH);  // Sample Pulse
     delayMicroseconds(1);
     digitalWrite(SEL1,LOW);
    
     delayMicroseconds(40); // 30
  
     digitalWrite(SEL2,LOW);

     //delayMicroseconds(250);

   // Wait for End of Measure
     while(micros()-time < 1000)
     {
        delayMicroseconds(10); 
     } 

//#define DBG2
long T[NMEAS];

noInterrupts();

for(int i=0;i<NMEAS;i++) 
{
  T[i]=Ts[i];
  if(T[i]<0 || T[i] > 10000 ) T[i]=0;
}
interrupts();

  value=analyze2(T+1,meas+ch); // 0-ter Wert in T ist die Startzeit der Messung, daher T+1
  if(value==0)
  {
    // Fallback, falls noch kein Quartett gemessen wurde:
    value=analyze0(T+1,meas+ch); // 0-ter Wert in T ist die Startzeit der Messung, daher T+1
  }





 
if(debugmode==1)
{   Serial.print(" VAL:");
    Serial.print(value);
 

    for( int i=1;i<count;i++)
    {
      Serial.print(" X"); Serial.print(i); Serial.print(":"); Serial.print(T[i]);
    }
    Serial.println();
} 
}
while(value==0);
  valold = 0.333 * (double) value + 0.666* valold;
  value=valold;
  return value;
}
#endif


#ifdef MOTORSHIELD
/*
 * PWM value to motor
 */
void pwmOut(int i,int out) {     
  if (out > 0) {                         // if REV > encoderValue motor move in forward direction.    
     switch(i)
     {
      case 0:     
       digitalWrite(MOTOR_A_DIR_PIN, HIGH); //Establishes forward direction of Channel A
       digitalWrite(MOTOR_A_BRAKE_PIN, LOW);   //Disengage the Brake for Channel A
       analogWrite(MOTOR_A_PWM_PIN, out);   //Spins the motor on Channel A at full speed
      break;
      case 1:
       digitalWrite(MOTOR_B_DIR_PIN, HIGH); //Establishes forward direction of Channel A
       digitalWrite(MOTOR_B_BRAKE_PIN, LOW);   //Disengage the Brake for Channel A
       analogWrite(MOTOR_B_PWM_PIN, out);   //Spins the motor on Channel A at full speed
      break;
     }
  }
  else {
     switch(i) {
      case 0:
       digitalWrite(MOTOR_A_DIR_PIN, LOW); //Establishes forward direction of Channel A
       digitalWrite(MOTOR_A_BRAKE_PIN, LOW);   //Disengage the Brake for Channel A
       analogWrite(MOTOR_A_PWM_PIN, abs(out));   //Spins the motor on Channel A at full speed
      break;
      case 1:
       digitalWrite(MOTOR_B_DIR_PIN, LOW); //Establishes forward direction of Channel A
       digitalWrite(MOTOR_B_BRAKE_PIN, LOW);   //Disengage the Brake for Channel A
       analogWrite(MOTOR_B_PWM_PIN, abs(out));   //Spins the motor on Channel A at full speed
      break;
     }
  }
}
#endif


#ifdef MOTORSHIELD_V2
/*
 * PWM value to motor
 */
void pwmOut(int i,int out) {  
  int absout=out;
  if( absout<0) absout = -out;   

  
  if (out > 0)                          // if REV > encoderValue motor move in forward direction.    
      myMotor[i]->run(FORWARD);
  else
      myMotor[i]->run(BACKWARD);
  myMotor[i]->setSpeed(absout);
  // turn on motor
  //myMotor[i]->run(RELEASE);
}
#endif


void setup() {
  // put your setup code here, to run once:
  Serial.begin(38400); //initialize serial comunication 

  Serial.println("Setup Pulser");
#ifdef MOTORSHIELD
  pinMode(MOTOR_A_DIR_PIN, OUTPUT);
  pinMode(MOTOR_A_PWM_PIN, OUTPUT);
  pinMode(MOTOR_A_BRAKE_PIN, OUTPUT);
    // Release the breaks... Just to be sure
  digitalWrite(MOTOR_A_BRAKE_PIN,LOW);
   // digitalWrite(MOTOR_B_BRAKE_PIN,LOW);
 //Set direction for both motors
  digitalWrite(MOTOR_A_DIR_PIN, LOW);
    analogWrite(MOTOR_A_PWM_PIN, 0);

  pinMode(MOTOR_B_DIR_PIN, OUTPUT);
  pinMode(MOTOR_B_PWM_PIN, OUTPUT);
  pinMode(MOTOR_B_BRAKE_PIN, OUTPUT);
 
    // Release the breaks... Just to be sure
  digitalWrite(MOTOR_B_BRAKE_PIN,LOW);
   // digitalWrite(MOTOR_B_BRAKE_PIN,LOW);
 //Set direction for both motors
  digitalWrite(MOTOR_B_DIR_PIN, LOW);
  analogWrite(MOTOR_B_PWM_PIN, 0); 
  pwmOut(0,0);
  pwmOut(1,0);
#endif
#ifdef MOTORSHIELD_V2
  if (!AFMS.begin()) {         // create with the default frequency 1.6KHz
  // if (!AFMS.begin(1000)) {  // OR with a different frequency, say 1KHz
    Serial.println("Could not find Motor Shield. Check wiring.");
    while (1);
  }
  Serial.println("Motor Shield found.");
  delay(1000);
#endif


   pinMode(SEL1, OUTPUT); // Excitation / Trigger 
   pinMode(SEL2, OUTPUT); // Sample Gate
#ifdef TTLCOUNTER
   pinMode(CTRA, OUTPUT); // Nibble Select S1
   pinMode(CTRB, OUTPUT); // Nibble Select S0

   pinMode(CTR1, INPUT);  // Counter Bit 0 
   pinMode(CTR2, INPUT);  // Counter Bit 1
   pinMode(CTR3, INPUT);  // Counter Bit 2
 #endif

#ifdef TIMER1MEAS
   setupTimer1();
#endif

   // Channel X or Y
  pinMode(CHSEL,OUTPUT);
  digitalWrite(CHSEL,LOW);
  
 // pinMode(CTL,OUTPUT);
 // digitalWrite(CTL,LOW);
  
  digitalWrite(SEL2,LOW);
  digitalWrite(SEL1,LOW);

  myPID0.SetMode(AUTOMATIC);   //set PID in Auto mode
  myPID0.SetSampleTime(1);  // refresh rate of PID controller
  myPID0.SetOutputLimits(-ulimit, ulimit); // this is the MAX PWM value to move motor, here change in value reflect change in speed of motor.
  myPID1.SetMode(AUTOMATIC);   //set PID in Auto mode
  myPID1.SetSampleTime(1);  // refresh rate of PID controller
  myPID1.SetOutputLimits(-ulimit, ulimit); // this is the MAX PWM value to move motor, here change in value reflect change in speed of motor.

  
  setupMeasurement( meas , channels0,DT0,TM0);
  setupMeasurement( meas+1 , channels1,DT1,TM1);
  Serial.println("done");
}

float v0_avg0,v1_avg;

float average(float v, float * v_avg, int k)
{
   *v_avg = *v_avg + (v- *v_avg)/(float)k;
   return *v_avg;
}

#define V1 70
#define V2 250
#define V3 480

int linearize(int v, int * v0, int * v1)
{
  double a [6][4]= 
{ 
   { 4.56833147e-04, -4.83837433e-02,  2.99635332e+00,  9.99563341e+01},
   { 1.09603739e-05, -1.60061195e-03,  2.77576348e-01,  2.66997234e+01},
   { 7.04689031e-06, -4.74717873e-03,  1.47752642e+00, -1.40339681e+02},
   {-4.05035683e-04, -4.72786162e-02, -2.66979827e+00,  1.01748946e+02},
   {-1.06278037e-05, -1.93872939e-03 ,-3.71604694e-01,  2.45836269e+01},
   {-4.15596665e-06, -1.55065931e-03, -3.38838321e-01, -9.46024817e+00}
};
   int r= -1;
   
   if(v<0 && v> -V1) {r=3;   }
   else if( v>= -V2 && v<= -V1) { r=4;  }
   else if( v< -V2 && v>= -V3) { r=5;  v -=50;}
   else if( v>=0 && v <V1) { r=0;  }
   else if( v>=V1 && v <=V2) { r=1;  }
   else if( v>V2 && v <=V3) { r=2; v+=50; }
  
   *v0 = (int) (a[r][3] + v*(a[r][2] + v*(a[r][1] + v*a[r][0])));
   //*v0 = (int) (a[r][3] + v*a[r][2] + v*v*a[r][1] + v*v*v*a[r][0]);
   if( v<0)
      *v0 = - *v0;
      /*
        Serial.print(" v=");
      Serial.print(v);
      Serial.print(" v0=");
            Serial.print(*v0);
                  Serial.print(" v1=");
                        Serial.print(*v1);
                         Serial.print(" r=");
      Serial.println(r);
      */
   switch(r)
   {
     case 0:  *v1= -80; break;
     case 1:  *v1= 0;   break;
     case 2:  *v0 *=70; *v0 /=100; *v1= *v0; break;
     case 3:  *v1= 80;  break;
     case 4:  *v1= 0;   break;
     case 5:  *v0 *=70; *v0 /=100; *v1= *v0; break;
   }
}

int calcPWMs( int v, int * pwm0, int * pwm1)
{
  int sgn=1;
  if(v<0)
  {
    sgn= -1;
    v= -v;
  }
  if(v<2)
  {
    *pwm0=0;
    *pwm1=0;
  } else if(v<=80)
    { 
      *pwm1= -80 * sgn;
      *pwm0 = sgn * ((int) ((float)(v) * (float) 250 / (float) 80) + 80);
    } else if(v<=250)
    {
      *pwm1=0;
      *pwm0 = sgn * ((int) ((float) (v-80) * (float) (170-50) / (float)(250-80)) + 50);
    } else
    {
      *pwm1 = *pwm0 = sgn *((int) ( (float)(v-250) / (float) (480-250) * (float) (255-50)) + 50);
      if( *pwm0 > 255) { *pwm1= sgn* 255; *pwm0= sgn*255; }
    }
  }
  
/*
void printHelp()
{
  Serial.println("Magnetostrictive Position Scanner and optional Motor Control");
  Serial.println("Commands:");
  Serial.println("h               - Help");
  Serial.println("p               - Toggle print mode, Prints can be visualized with Arduino-Plotter");
  Serial.println("p               - Toggle debug mode, Prints Raw Data and Results of analyze function ");
  Serial.println("                  Can be used to calculate DT and TM values ");
  Serial.println("Motor Control commands:");
  Serial.println("Motors are controlled with Adafrout V2 i2C Motor Shield");
  Serial.println("s               - Start / Stop Motor");
  Serial.println("m<pos>,<pos2>   - Move to position (<pos1>,<pos2>)");
  delay(2000);
}
*/

String readString; //This while store the user input data
volatile long old_time=0;


volatile long new_time;

long input1_old;
  long dt;
  long dcount;
  long t_old;
  double setpoint0old, setpoint1old;
float d0_old,d1_old;
  /*  
 *   State Values 
 */
#define OK      0
#define MOVING  1
#define ERR    99
int state=OK;

void loop() {
  char buf[128];
  char * ptr;
  float v;
  int ch=1;
  long t;
  int err0=0,err1=0;
  float delta1;
  float delta0;
  float delta;
  float dt;
  float d0,d1;
    
  
  // put your main code here, to run repeatedly:

  // put your main code here, to run repeatedly:
 while (Serial.available()) { //Check if the serial data is available.
    delay(3);                  // a small delay
    char c = Serial.read();  // storing input data
    readString += c;         // accumulate each of the characters in readString
  }
  
 if (readString.length() >0) { //Verify that the variable contains information  
    readString.toCharArray(buf, 100);
    // Toggle motor on/off
    if (strncmp(buf,"s",1)==0)
    {
      Serial.print("Runmode ");   
      setpoint1=measure(1);
      setpoint0=measure(0);
      
      
      runmode=1-runmode;
      if(runmode==0)
      {
      pwmOut(0,0);   pwmOut(1,0);   pwmOut(2,0);   pwmOut(3,0);
      }
      Serial.println(runmode);
      delay(1000);
    }
      if (strncmp(buf,"d",1)==0)
    {
      Serial.print("Debugmode ");   
      
      
      
      debugmode=1-debugmode;
     
      Serial.println(debugmode);
      delay(1000);
    }
    // Toggle print mode
    if (strncmp(buf,"p",1)==0)
    {
      Serial.print("Printmode "); 
      printmode=1-printmode;
      Serial.println(printmode);
      delay(1000);
    }
   /* if (strncmp(buf,"h",1)==0)
    { 
      printHelp();
    }*/
     if (strncmp(buf,"M",1)==0)
    { 
      ptr = strtok(buf+1, ",");
      int i=atoi(ptr);
      Serial.print("Motor "); Serial.println(i);
      pwmOut(i,255);
      delay(1000);
      pwmOut(i,-255);
      delay(1000);
      pwmOut(i,0);
    }
    // Move motor
    if (strncmp(buf,"m",1)==0)
    { 
      ptr = strtok(buf+1, ",");
      float x=atof(ptr);
     
      ptr = strtok(NULL, ",");
      if(ptr !=NULL)
      {   
        float y=atof(ptr);
        setpoint0=x;
        setpoint1= y;
        
        state=MOVING;
      }
    }  
     // Velocities to motor
    if (strncmp(buf,"V",1)==0)
    { 
      long t0,t1;
      ptr = strtok(buf+1, ",");
      int v=atoi(ptr);
      int pwm0,pwm1;
      for(v= 0;v<=450;v+=10)
      {
      linearize(v,&pwm0,&pwm1);
     /*  Serial.print("v="); Serial.print(v);
      Serial.print(" PWM0="); Serial.print(pwm0);
      Serial.print(" PWM1="); Serial.print(pwm1);
      Serial.println();*/

      long p_0 = measure(ch);
      pwmOut(0+ch*2,pwm0);
      pwmOut(1+ch*2,pwm1);
      
      long p_prev = p_0;   
      t0=millis();
      long t1_prev=t0;
      long vcum=0;
      do {
      do {
        t1=millis();
      } while( (t1-t1_prev)<100);
      long p_act = measure(ch); 
      float v_act = (p_act - p_prev);
      vcum += v_act;
      t1_prev=t1;
      p_prev=p_act;
      average( v_act, &v1_avg,4);
      if(t1-t0>2000)
      {
    pwmOut(0+ch*2,0);
      pwmOut(1+ch*2,0);
      }
      Serial.print(t1-t0); Serial.print(", "); Serial.print(v_act); Serial.print(", ");  Serial.print(v1_avg); Serial.print(", ");  Serial.println(vcum);
      } while(t1-t0<2400);
      
      //delay(2000);
 
      //delay(400);
      long p_1 = measure(ch);
     
      pwmOut(0+ch*2,-pwm0);
      pwmOut(1+ch*2,-pwm1);
        delay(2000);
        pwmOut(0+ch*2,0);
        pwmOut(1+ch*2,0);
        delay(400);
         long p_2 = measure(ch);
        
      //  Serial.print(v); Serial.print(", "); Serial.print(pwm0);  Serial.print(", "); Serial.print(pwm1); Serial.print(", "); Serial.print( p_1 - p_0); Serial.print(", ");
      //  Serial.println( p_2 - p_1);        
         }
    }
      
    // Velocities to motor
    if (strncmp(buf,"v",1)==0)
    { 
      ptr = strtok(buf+1, ",");
      float v0;
      int ch=atoi(ptr);
     
      ptr = strtok(NULL, ",");
      if(ptr !=NULL)
      { 
        float v1=atof(ptr);  
        for(v0=20;v0<=250;v0+=10)
        {
        long p_0 = measure(ch);
        pwmOut(0+ch*2,v0);
        pwmOut(1+ch*2,v1);
        delay(2000);
        pwmOut(0+ch*2,0);
        pwmOut(1+ch*2,0);
        delay(400);
        long p_1 = measure(ch);
        pwmOut(0+ch*2,-v0);
        pwmOut(1+ch*2,-v1);
        delay(2000);
        pwmOut(0+ch*2,0);
        pwmOut(1+ch*2,0);
        delay(400);
         long p_2 = measure(ch);
        Serial.print(v0); Serial.print(", "); Serial.print(v1); Serial.print(", "); Serial.print( p_1 - p_0); Serial.print(", ");
        Serial.println( p_2 - p_1);        
        }
      
      } }
     readString="";
    }

#ifdef DIFFERENTIAL
    d0=measure(0);
    d1=measure(1);
    t=micros();
    dt = t-t_old;
    double  v1 = 1000 * (d1-d1_old)  / ( dt );
    double v0 = 1000 * (d0 -d0_old) /  ( dt );
    d0_old=d0;
    d1_old=d1;
    t_old=t;
 
    if(setpoint0==0) setpoint0=d0;
    if(setpoint1==0) setpoint1=d1;
    if(abs(setpoint0 - d0) < 5 && abs(setpoint1 - d1)< 5) 
{
   state=OK;
 
   pwmOut(0,0);
   pwmOut(1,0);
   pwmOut(2,0);
   pwmOut(3,0);
  // digitalWrite(MOTOR_A_BRAKE_PIN,HIGH);
  // digitalWrite(MOTOR_B_BRAKE_PIN,HIGH);
   setpoint0old=setpoint0;
   setpoint1old=setpoint1;
   
}
else
{
 //  digitalWrite(MOTOR_A_BRAKE_PIN,LOW);
 //  digitalWrite(MOTOR_B_BRAKE_PIN,LOW);
}


   delta1 = (setpoint1-d1);
   delta0 = (setpoint0-d0);
   delta = sqrt( delta1*delta1 + delta0*delta0);
  input0 = delta;
  double tmp=setpoint0;
  setpoint0 = 0.0;
     myPID0.Compute();   
  setpoint0=tmp;
    

  if(delta>0)
  {
  double v0soll = output0 * delta0 / delta;
  double v1soll = output0 * delta1 / delta ;
  output0 = -v0soll;
  output1 = -v1soll;

 /*
  double delta_target0 = setpoint0-setpoint0old;
  double delta_target1 = setpoint1-setpoint1old;
  double delta_target2 = delta_target0*delta_target0 + delta_target1*delta_target1;
  if (delta_target2 != 0.0)
  {
  double z = delta_target0*delta0 + delta_target1*delta1;
  double rho0 = z / delta_target2 * delta_target0;
  double rho1 = z / delta_target2 * delta_target1;
  double eps0 = rho0 - delta0;
  double eps1 = rho1 - delta1;

  //output0 /= 1.3;

  output0 -= 6*eps0 ;
  output1 -= 6*eps1 ;
  */
  /*
  Serial.print(d0);Serial.print(", "); 
  Serial.print(output0);Serial.print(", "); 
  Serial.print(eps0); Serial.print(", "); 
   Serial.print(d1);Serial.print(", "); 
  Serial.print(output1);Serial.print(", "); 
  Serial.println(eps1); 
  */
  //}
  // Serial.print(-v0soll); Serial.print(" "); Serial.print(output0); Serial.print(" "); Serial.print(-v1soll); Serial.print(" "); Serial.println(output1);
  }
  else
  {
    output0=output1=0;
  }
  int pwm[4];
    linearize(output0,pwm,pwm+1);

    linearize(output1,pwm+2,pwm+3);
    
    if(runmode==1 && state==MOVING)
    {
     // if(abs( setpoint1-input1 ) < 3)
     // {
     //   pwmOut(1+ch*2,0);
     //   pwmOut(0+ch*2,0);
    //  }
    //  else
    //  {
     for(int mych=0;mych<2;mych++)
     {
       pwmOut(1+mych*2,pwm[mych*2+1]);
       pwmOut(0+mych*2,pwm[mych*2]);
      }
    }
 #else
    d0=measure(0);
    d1=measure(1);
    t=micros();
    dt = t-t_old;
    double  v1 = 1000 * (d1-d1_old)  / ( dt );
    double v0 = 1000 * (d0 -d0_old) /  ( dt );
   
    t_old=t;
 
    if(setpoint0==0) setpoint0=d0;
    if(setpoint1==0) setpoint1=d1;
    if(abs(setpoint0 - d0) < 5 && abs(setpoint1 - d1)< 5) 
{
   state=OK;
 
   pwmOut(0,0);
   pwmOut(1,0);
   pwmOut(2,0);
   pwmOut(3,0);
  // digitalWrite(MOTOR_A_BRAKE_PIN,HIGH);
  // digitalWrite(MOTOR_B_BRAKE_PIN,HIGH);
   //setpoint0old=setpoint0;
   //setpoint1old=setpoint1;
   
}
else if (abs(setpoint0 - d0) > 10 || abs(setpoint1 - d1)>10) 
{
  state=MOVING;
}
    input0=d0;
    input1=d1;
    myPID0.Compute();   
    myPID1.Compute();  
    if(runmode==1) // && state==MOVING)
    {
      if(output0>0 && output0<llimit) output0=llimit; 
      if(output0<0 && output0>-llimit) output0= - llimit;
      if(output1>0 && output1<llimit) output1=llimit; 
      if(output1<0 && output1>-llimit) output1= - llimit;
      
       pwmOut(0,output0);   
       pwmOut(1,output1);   
    }
 #endif
    
    if(printmode)
    {
    Serial.print(" S0:");
    Serial.print(setpoint0);
    Serial.print(" S1:");
    Serial.print(setpoint1);
    Serial.print(" C0:");
    Serial.print(d0);
    Serial.print(" o0:");
    Serial.print(output0);
    Serial.print(" C1:");
    Serial.print(d1);
    Serial.print(" o1:");
    Serial.print(output1);
    Serial.print(" MIN:"); Serial.print(0);
    Serial.print(" MAX:"); 
    if(err0==1) Serial.print("*"); else Serial.print(".");

    if(err1==1) Serial.print("*"); else Serial.print(".");
   Serial.print(2000);
   
   Serial.println();
    }
 }
